#include "mv.hh"
#include <assert.h>
#include <cmath>
#include <iostream>
#include <stdlib.h> // for rand() function

using namespace std ;

// initialize structure to a matrix of size 0 x 0
template <typename valueType>
Mat<valueType>::Mat( ) {
  cout << "Call constructor for Mat\n" ;
  _nrow        = 0 ;
  _ncol        = 0 ;
  memory_block = nullptr ;
  values       = nullptr ;
}

template <typename valueType>
Mat<valueType>::Mat( int _nr, int _nc ) : _nrow(_nr), _ncol(_nc) {
  cout << "Call constructor for Mat with size = "
       << _nr << ", " << _nc << "\n" ;

  memory_block = new valueType [_nrow*_ncol] ;
  values       = new valueType* [_nrow] ;

  // initialize the pointer of pointer
  for ( int i = 0 ; i < _nrow ; ++i )
    values[i] = memory_block + i * _ncol ;
}

// copy constructor
template <typename valueType>
Mat<valueType>::Mat( Mat<valueType> const & A ) {
  cout << "Mat: Call copy constructor\n" ;
  allocate(A.nrow(),A.ncol()) ;
  for ( int i = 0 ; i < A.nrow() ; ++i )
    for ( int j = 0 ; j < A.ncol() ; ++j )
      values[i][j] = A(i,j) ;
}

// copy operator
template <typename valueType>
Mat<valueType> const & 
Mat<valueType>::operator = ( Mat<valueType> const & A ) {
  cout << "Mat: Call copy operator\n" ;
  allocate(A.nrow(),A.ncol()) ;
  for ( int i = 0 ; i < A.nrow() ; ++i )
    for ( int j = 0 ; j < A.ncol() ; ++j )
      values[i][j] = A(i,j) ;
  return *this ;
}

template <typename valueType>
Mat<valueType>::~Mat() {
  cout << "Call destructor for Mat\n" ;
  // check if alreay de-allocated
  if ( memory_block != nullptr ) delete [] memory_block ;
  if ( values       != nullptr ) delete [] values ;
}

// alocate a matrix to the size: nrow x ncol
template <typename valueType>
void
Mat<valueType>::allocate( int nrow, int ncol ) { // USE this to distinguish arguments
  // (*this).nrow ==> this -> nrow
  if ( _nrow == nrow && _ncol == ncol ) return ; // nothing to do!

  _nrow = nrow ;
  _ncol = ncol ;

  if ( memory_block != nullptr ) delete [] memory_block ;
  memory_block = new valueType [nrow*ncol] ;

  if ( values != nullptr ) delete [] values ;
  values = new valueType* [_nrow] ;

  // initialize the pointer of pointer
  for ( int i = 0 ; i < _nrow ; ++i )
    values[i] = memory_block + i * _ncol ;
}

template <typename valueType>
void
Mat<valueType>::fill_random() {
  for ( int i = 0 ; i < _nrow ; ++i )
    for ( int j = 0 ; j < _ncol ; ++j )
      values[i][j] = (rand() - rand())/double(RAND_MAX) ;
}

template <typename valueType>
void
Mat<valueType>::fill( valueType val ) {
  for ( int i = 0 ; i < _nrow ; ++i )
    for ( int j = 0 ; j < _ncol ; ++j )
      values[i][j] = val ;
}

/*
    _    _            _               
   / \  | | __ _  ___| |__  _ __ __ _ 
  / _ \ | |/ _` |/ _ \ '_ \| '__/ _` |
 / ___ \| | (_| |  __/ |_) | | | (_| |
/_/   \_\_|\__, |\___|_.__/|_|  \__,_|
           |___/                      
*/

template <typename valueType>
Mat<valueType> operator + ( Mat<valueType> const & A, Mat<valueType> const & B ) {
  assert( A.nrow() == B.nrow() ) ;
  assert( A.ncol() == B.ncol() ) ;
  Mat<valueType> res(A.nrow(),A.ncol()) ;
  for ( int i = 0 ; i < A.nrow() ; ++i )
    for ( int j = 0 ; j < A.ncol() ; ++j )
      res(i,j) = A(i,j) + B(i,j) ;
  return res ;
}

template <typename valueType>
Mat<valueType> operator - ( Mat<valueType> const & A, Mat<valueType> const & B ) ;

template <typename valueType>
Mat<valueType> operator * ( Mat<valueType> const & A, Mat<valueType> const & B ) ;

template <typename valueType>
Mat<valueType> operator * ( double a, Mat<valueType> const & B ) ;

template <typename valueType>
Mat<valueType> operator * ( Mat<valueType> const & A, valueType b ) ;

template <typename valueType>
Mat<valueType> operator / ( Mat<valueType> const & A, valueType b ) ;

template <typename valueType>
void
Mat_x_Vec( struct Mat<valueType> const & mat, // ensure the compiler that mat is not modified inside...
           struct Vec<valueType> const & vec,
           struct Vec<valueType>       & res ) {
  // check matrix vector compatibility
  if ( mat.ncol() != vec.size() ) {
    cerr << "Mat_x_Vec, incompatible dimensions\n" ;
    exit(0) ;
  }   
  if ( res.size() != mat.nrow() ) res.allocate( mat.nrow() ) ;
  for ( int i = 0 ; i < mat.nrow() ; ++i ) {
    res(i) = 0 ;
    for ( int j = 0 ; j < mat.ncol() ; ++i )
      res(i) += mat(i,j)*vec(j) ;
  }
}

template <typename valueType>
void
Mat_x_Mat( struct Mat<valueType> const & A, 
           struct Mat<valueType> const & B,
           struct Mat<valueType>       & C ) {
  // check compatibility
  if ( A.ncol() != B.nrow() ) {
    cerr << "Mat_x_Mat_123, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.allocate( A.nrow(), B.ncol() ) ;

  // set to 0
  C.fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int i = 0 ; i < A.nrow() ; ++i ) // 1
    for ( int j = 0 ; j < B.ncol() ; ++j ) // 2
      for ( int k = 0 ; k < A.ncol() ; ++k ) // 3
        C(i,j) += A(i,k)*B(k,j) ;
}

/*
 ___    _____  
|_ _|  / / _ \ 
 | |  / / | | |
 | | / /| |_| |
|___/_/  \___/ 

*/

template <typename valueType>
ostream &
operator << ( ostream & stream, Mat<valueType> const & mat ) {
  for ( int i = 0 ; i < mat.nrow() ; ++i ) {
    stream << mat(i,0) ;
    for ( int j = 1 ; j < mat.ncol() ; ++j )
      stream << " " << mat(i,j) ;
    stream << '\n' ;
  }
  return stream ;
}

template class Mat<float> ; // force instantiation of all the methods of class Mat<float>
template class Mat<double> ; // force instantiation of all the methods of class Mat<double>
template class Mat<int> ; // force instantiation of all the methods of class Mat<int>

template ostream & operator << ( ostream & stream, Mat<float> const & mat ) ;
template ostream & operator << ( ostream & stream, Mat<double> const & mat ) ;
template ostream & operator << ( ostream & stream, Mat<int> const & mat ) ;
